"""Handlers for ensuring the mock data passed will work."""
import types

from .exceptions import UnevenColumnError, ArgError, OptionError


def _check_col_length(datasets):
    """
    Ensures the length of all data columns are equal within each
    dataset.

    Parameters
    ----------
    datasets: dict[]
        List of datasets to validate.

    Raises
    ------
    UnevenColumnError
        If the columns do not equal each other.
    """
    for dataset in datasets:
        data = dataset['data']
        col_length = len(data[0])
        for row in data:
            if len(row) != col_length:
                raise UnevenColumnError(data)


def _check_arg_consistency(datasets):
    """
    Ensures if arg is specified on one dataset, it is specified on all
    datasets. Also checks that the same arg is not reused for multiple
    csv assignments.

    Parameters
    ----------
    datasets: dict[]
        List of dicts containing the datasets.

    Raises
    ------
    ArgError
        If arg is specified on one dataset and not all others.
    """
    specified = False
    assigned_args = []
    for dataset in datasets:
        if dataset.get('arg'):
            if dataset['arg'] in assigned_args:
                raise ArgError(f'arg {dataset["arg"]} already assigned.')
            assigned_args.append(dataset['arg'])
            specified = True
        if specified and not dataset.get('arg'):
            raise ArgError


def _check_dataset_options(datasets):
    """
    Ensures that for each dataset, if options are provided that they
    are valid.

    Parameters
    ----------
    datasets: dict[]
        List of dicts containing the datasets.

    Raises
    ------
    OptionError
        If options do not meet the requirements.
    """
    for dataset in datasets:
        if not dataset.get('opts'):
            continue

        valid_options = [
            'row_pattern',
            'add_rows',
            'base_row_index',
            'increment'
        ]
        if not all(val in list(dataset['opts'].keys()) for val in valid_options[:2]):  # noqa
            raise OptionError(f'Options must include {valid_options[:2]}')
        for key in dataset['opts'].keys():
            if key not in valid_options:
                raise OptionError(f'Invalid option {key}.')  # noqa
        if (dataset['opts']['row_pattern'] not in ['copy']
                and
                not callable(dataset['opts']['row_pattern'])):
            raise OptionError(f'row_pattern option must be one of [\'copy\'] or a function')  # noqa


def validate(datasets):
    """
    Helper function to apply all validations.

    Parameters
    ----------
    datasets: dict[]
        list of datasets to validate.
    """
    _check_col_length(datasets)
    _check_arg_consistency(datasets)
    _check_dataset_options(datasets)



"""
Handling of options passed to the decorator to generate more rows of data.
"""


def _copy_rows(data, num_rows, base_row_index=0):
    """
    Copies one list from a list of lists and adds it to the same list n times.

    Parameters
    ----------
    data: [][]
        Dataset to apply the function to. This will be mutated.
    base_row_index: int
        The index of the row to copy
    num_rows: int
        The number of times to copy the base row.

    Returns
    -------
    [][]
        List of lists, containing the additional copied data.
    """
    copies = [data[base_row_index] for _ in range(num_rows)]
    return data + copies


def _apply_func(data, func, num_rows, base_row_index=0, increment=False):
    """
    Apply the function to the base row which returns a new row.
    This is then added to the dataset n times.

    Parameters
    ----------
    data: [][]
        List to apply the function to.
    func: function
        The function to apply to the row. This won't alter the initial row
        it is applied to.
    base_row_index: int
        The index of the row to first apply the function to.
    num_rows: int
        The number of times this function should be applied. Will result in
        this many new rows added to the dataset.
    increment: boolean
        If true, the function will be applied to the newly created rows rather
        than the base row on further iterations.

    Returns
    -------
    [][]
        The mutated list with the new rows added.
    """
    row = list(data[base_row_index])
    curr_index = base_row_index
    for _ in range(num_rows):
        data.append(func(row))
        if increment:
            curr_index += 1
            row = list(data[curr_index])
    return data


def apply_options(data, opts):
    """
    Applies the passed options to the dataset.

    Parameters
    ----------
    data: [][]
        List of lists containing the mock csv data.
    opts: dict
        Dictionary of options used to determine how to mutate the data.
        Options are as follows:
            add_rows: int
                The number of rows to add onto the mock dataset.
            row_pattern: str or function
                if str, should be 'copy' to apply the copy function.
                if function, will be used to create the new rows.
            base_row_index: int
                The index of the base row to apply the row pattern to.
            increment: boolean,
                If true and row_pattern is a function, the function will
                be applied incrementally on the dataset, rather than just
                on the base row over and over again.

    Returns
    -------
    [][]
        The dataset with the new rows added.
    """
    if not opts:
        return data
    if opts['row_pattern'] == 'copy':
        data = _copy_rows(
            data=data,
            num_rows=opts['add_rows'],
            base_row_index=opts.get('base_row_index') or 0
        )
    else:
        data = _apply_func(
            data=data,
            func=opts['row_pattern'],
            num_rows=opts['add_rows'],
            base_row_index=opts.get('base_row_index') or 0,
            increment=opts.get('increment') or False
        )
    return data



import unittest

from quikcsv import options as opts


class Test_Copy_Rows(unittest.TestCase):

    def test_valid_input(self):
        data = [
            ['a', 'b', 'c'],
            ['d', 'e', 'f']
        ]
        expected = [
            ['a', 'b', 'c'],
            ['d', 'e', 'f'],
            ['d', 'e', 'f'],
            ['d', 'e', 'f']
        ]
        output = opts._copy_rows(data, 2, 1)
        self.assertEqual(output, expected)

    def test_bad_base_index(self):
        data = [['a', 'b', 'c']]
        with self.assertRaises(IndexError):
            opts._copy_rows(data, 1, 2)


class Test_Apply_Func(unittest.TestCase):

    def test_valid_input_no_increment(self):
        data = [['a', 'b', 'c']]
        expected = [
            ['a', 'b', 'c'],
            ['aa', 'ba', 'ca'],
            ['aa', 'ba', 'ca'],
        ]
        output = opts._apply_func(
            data,
            lambda x: [i + 'a' for i in x],
            2
        )
        self.assertEqual(expected, output)

    def test_valid_input_with_increment(self):
        data = [['a', 'b', 'c']]
        expected = [
            ['a', 'b', 'c'],
            ['aa', 'ba', 'ca'],
            ['aaa', 'baa', 'caa']
        ]

        output = opts._apply_func(
            data=data,
            func=lambda x: [i + 'a' for i in x],
            num_rows=2,
            increment=True
        )
        self.assertEqual(expected, output)


class Test_Apply_Options(unittest.TestCase):

    def test_with_copy_pattern(self):
        data = [['a', 'b', 'c']]
        expected = [
            ['a', 'b', 'c'],
            ['a', 'b', 'c'],
            ['a', 'b', 'c']
        ]
        output = opts.apply_options(
            data,
            opts=dict(
                add_rows=2,
                row_pattern='copy'
            )
        )
        self.assertEqual(expected, output)

    def test_with_custom_func_pattern(self):
        data = [['a', 'b', 'c']]
        expected = [
            ['a', 'b', 'c'],
            ['aa', 'ba', 'ca'],
            ['aa', 'ba', 'ca']
        ]
        output = opts.apply_options(
            data,
            opts=dict(
                add_rows=2,
                row_pattern=lambda x: [i + 'a' for i in x]
            )
        )
        self.assertEqual(expected, output)

    def test_with_custom_func_pattern_and_increment(self):
        data = [['a', 'b', 'c']]
        expected = [
            ['a', 'b', 'c'],
            ['aa', 'ba', 'ca'],
            ['aaa', 'baa', 'caa']
        ]
        output = opts.apply_options(
            data,
            opts=dict(
                add_rows=2,
                row_pattern=lambda x: [i + 'a' for i in x],
                increment=True
            )
        )
        self.assertEqual(expected, output)



import unittest

from quikcsv import validation, exceptions


class Test_Column_Length_Validation(unittest.TestCase):

    def test_even_columns_should_not_raise_error(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ]
            )
        ]
        validation._check_col_length(datasets)

    def test_uneven_columns_should_raise_error(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b'],
                    ['a', 'b', 'c']
                ]
            )
        ]
        with self.assertRaises(exceptions.UnevenColumnError):
            validation._check_col_length(datasets)

    def test_two_datasets(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ]
            ),
            dict(
                data=[
                    ['a', 'b', 'c', 'd'],
                    ['a', 'b', 'c', 'd'],
                ]
            )
        ]
        validation._check_col_length(datasets)


class Test_Arg_Consistency(unittest.TestCase):

    def test_assigning_two_different_args_should_not_raise_error(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                arg='one'
            ),
            dict(
                data=[
                    ['a', 'b', 'c', 'd'],
                    ['a', 'b', 'c', 'd'],
                ],
                arg='two'
            )
        ]
        validation._check_arg_consistency(datasets)

    def test_assigning_same_arg_should_raise_error(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                arg='one'
            ),
            dict(
                data=[
                    ['a', 'b', 'c', 'd'],
                    ['a', 'b', 'c', 'd'],
                ],
                arg='one'
            )
        ]
        try:
            validation._check_arg_consistency(datasets)
            self.fail('ArgError not raised.')
        except exceptions.ArgError as e:
            self.assertEqual(e.message, 'arg one already assigned.')


class Test_Check_Dataset_Options(unittest.TestCase):

    def test_required_options_should_validate(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                opts={
                    'row_pattern': 'copy',
                    'add_rows': 2
                }
            )
        ]
        try:
            validation._check_dataset_options(datasets)
        except exceptions.OptionError:
            self.fail('Options did not validate.')

    def test_missing_required_options_should_not_validate(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                opts={
                    'row_pattern': 'copy'
                }
            )
        ]
        try:
            validation._check_dataset_options(datasets)
            self.fail('OptionsError wasn not raised.')
        except exceptions.OptionError as e:
            self.assertEqual(e.message, 'Options must include [\'row_pattern\', \
                                         \'add_rows\']')
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                opts={
                    'add_rows': 2
                }
            )
        ]
        try:
            validation._check_dataset_options(datasets)
            self.fail('OptionsError wasn not raised.')
        except exceptions.OptionError as e:
            self.assertEqual(e.message, 'Options must include [\'row_pattern\', \
                                         \'add_rows\']')

    def test_passing_invalid_options_should_not_validate(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                opts={
                    'row_pattern': 'copy',
                    'add_rows': 2,
                    'invalid': 0
                }
            )
        ]
        try:
            validation._check_dataset_options(datasets)
            self.fail('OptionsError was not raised.')
        except exceptions.OptionError as e:
            self.assertEqual(e.message, 'Invalid option invalid.')

    def test_row_pattern_option_valid_checks(self):
        datasets = [
            dict(
                data=[
                    ['a', 'b', 'c'],
                    ['a', 'b', 'c']
                ],
                opts={
                    'row_pattern': 'bad_pattern',
                    'add_rows': 2,
                }
            )
        ]
        try:
            validation._check_dataset_options(datasets)
            self.fail('OptionsError was not raised.')
        except exceptions.OptionError as e:
            self.assertEqual(e.message, 'row_pattern option must be one of \
                                         [\'copy\'] or a function')



import unittest
import csv

from quikcsv.decorator import QuikCSV


class Test_QuikCSV(unittest.TestCase):

    def test_single_dataset_data_only(self):
        datasets = [dict(
            data=[
                ['A', 'B', 'C'],
                ['1', '2', '3']
            ]
        )]

        @QuikCSV(datasets)
        def stub(csvs):
            reader = csv.reader(csvs[0])
            output = [row for row in reader]
            self.assertEqual(datasets[0]['data'], output)
        stub()

    def test_single_dataset_with_arg(self):
        datasets = [dict(
            data=[
                ['A', 'B', 'C'],
                ['1', '2', '3']
            ],
            arg='csv_'
        )]

        @QuikCSV(datasets)
        def stub(csv_):
            reader = csv.reader(csv_)
            output = [row for row in reader]
            self.assertEqual(datasets[0]['data'], output)
        stub()

        datasets[0]['arg'] = 'second'

        @QuikCSV(datasets)
        def stub_two(first, second):
            reader = csv.reader(second)
            output = [row for row in reader]
            self.assertEqual(datasets[0]['data'], output)
        stub_two(1)
        stub_two(1, 2)

    def test_single_dataset_with_opts_copy_pattern(self):
        datasets = [dict(
            data=[['a', 'b', 'c']],
            opts=dict(
                row_pattern='copy',
                add_rows=2
            )
        )]

        @QuikCSV(datasets)
        def stub(csvs):
            expected = [
                ['a', 'b', 'c'],
                ['a', 'b', 'c'],
                ['a', 'b', 'c']
            ]
            reader = csv.reader(csvs[0])
            output = [row for row in reader]
            self.assertEqual(expected, output)
        stub()

    def test_with_opts_func_pattern_no_increment(self):
        datasets = [dict(
            data=[['a', 'b', 'c']],
            opts=dict(
                row_pattern=lambda x: [i + 'a' for i in x],
                add_rows=2
            )
        )]

        @QuikCSV(datasets)
        def stub(csvs):
            expected = [
                ['a', 'b', 'c'],
                ['aa', 'ba', 'ca'],
                ['aa', 'ba', 'ca']
            ]
            reader = csv.reader(csvs[0])
            output = [row for row in reader]
            self.assertEqual(expected, output)
        stub()

    def test_with_opts_func_pattern_with_increment(self):
        datasets = [dict(
            data=[['a', 'b', 'c']],
            opts=dict(
                row_pattern=lambda x: [i + 'a' for i in x],
                add_rows=2,
                increment=True
            )
        )]

        @QuikCSV(datasets)
        def stub(csvs):
            expected = [
                ['a', 'b', 'c'],
                ['aa', 'ba', 'ca'],
                ['aaa', 'baa', 'caa']
            ]
            reader = csv.reader(csvs[0])
            output = [row for row in reader]
            self.assertEqual(expected, output)
        stub()



import csv
import io
from functools import wraps
import inspect

from .validation import validate
from .options import apply_options


def QuikCSV(datasets):
    """
    Decorator to quickly create temporary CSV mock files. These are passed to
    specified arguments in the decorated function (or to the first argument by
    default). These files are used like any other file in Python, but don't
    need to be opened or closed.

    Parameters
    ----------
    datasets: dict[]
        List of dictionaries which specificy the data to mock in the csv.
    """
    def outer(func):
        @wraps(func)
        def inner(*args, **kwargs):
            csvs = []
            validate(datasets)
            try:
                for dataset in datasets:
                    data_ = apply_options(
                        dataset['data'],
                        dataset.get('opts')
                    )
                    stream = io.StringIO()
                    writer = csv.writer(stream)
                    writer.writerows(data_)
                    stream.seek(0)
                    if dataset.get('arg'):
                        arg_names = inspect.getfullargspec(func).args
                        index = arg_names.index(dataset['arg'])
                        args = args[:index] + args[index+1:]
                        args = args[:index] + (stream,) + args[index:]
                    else:
                        csvs.append(stream)
                if len(csvs) > 0:
                    return func(csvs, *args, **kwargs)
                else:
                    return func(*args, **kwargs)
            finally:
                for stream in csvs:
                    stream.close()
        return inner
    return outer



import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='quikcsv',
    version='0.1',
    author='Bryn Mck',
    description='Decorator for making quick mock csv files for testing.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/PRIV00/quikcsv',
    packages=setuptools.find_packages(),
    python_required='>=3.8'
)





class Error(Exception):
    """Base Error"""
    pass


class UnevenColumnError(Error):

    def __init__(self, data, message="Mock data columns are uneven. Check \
                                      the length of your lists, they need to \
                                      match."):
        self.data = data
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.data} -> {self.message}'


class ArgError(Error):

    def __init__(self, message="arg must be specified on all datasets if \
                                specified on one."):
        self.message = message
        super().__init__(self.message)


class OptionError(Error):

    def __init__(self, message="Error with passed options."):
        self.message = message
        super().__init__(self.message)
